from osgeo import gdal, ogr, osr
import numpy as np
from scipy import stats
import math
import os


def raster_test():
    return "this is the raster module of the gispy package"


def addressOfCoordinates(x, y, geot):
    col = np.floor(np.divide(np.subtract(x, geot[0]), geot[1]))
    row = np.floor(np.divide(np.subtract(geot[3], y), abs(geot[5])))
    return row, col


def clipRasterBoundingBox(rasterpath, outputpath, bbox, bboxSrs=None, nodata=-9999, dstSrs=None, xres=None, yres=None,
                          resample=None):
    """
    Clip raster to coordinates of a bounding box.
    Args:
        rasterpath: raster to clip
        outputpath: path of output clipped raster
        bbox: extent of output raster (minX, minY, maxX, maxY)
        bboxSrs: SRS in which bbox coordinates are expressed, if different than SRS of input raster (default: None)
        nodata: nodata value (default: -9999)
        dstSrs: spatial reference for output raster, if different than input raster (default: None)
        xres: x resolution of output raster (default: None, resolution of input raster)
        yres: y resolution of output raster (default: None, resolution of input raster)
        resample: resampling algorithm. Example: gdal.GRA_Bilinear (default: None i.e. gdal.GRA_NearestNeighbor)

    Returns:

    """
    if xres is None or yres is None: xres, yres = getXYResolution(rasterpath)

    # to change resample algorithm gdal.GRA_Bilinear
    warpOptions = gdal.WarpOptions(format='GTiff', outputBounds=bbox, outputBoundsSRS=bboxSrs, xRes=xres,
                                   yRes=abs(yres), dstSRS=dstSrs, dstNodata=nodata, resampleAlg=resample)
    gdal.WarpOptions()
    gdal.Warp(outputpath, rasterpath, options=warpOptions)
    return None


def coordinatesOfAddress(row, col, geot):
    x = np.add(np.add(geot[0], np.multiply(col, geot[1])), 0.5*geot[1])
    y = np.add(np.add(geot[3], np.multiply(row, geot[5])), 0.5*geot[5])
    return (x, y)


def coordinatesOfLinearIndex(idx, geot, rows, cols, band=1):
    row = np.floor(np.divide(np.subtract(idx, rows*cols*(band-1)), cols))
    col = np.subtract(idx, np.add(np.multiply(row, cols), rows*cols*(band-1)))
    x, y = coordinatesOfAddress(row, col, geot)
    return x,y


def createBandIndex(rasterPath, minValue, maxValue):
    """

    Args:
        rasterPath:
        minValue:
        maxValue:

    Returns:

    """
    band = getRasterBandAsArray(rasterPath)
    if band is not None:
        array = band - minValue
        array = np.where((array < 0) | (array > (maxValue-minValue)), 0, array)
        return array
    else:
        return None


def createGDALRaster(filename, rows, cols, bands=1, datatype=gdal.GDT_Float32, drivername='GTiff', geot=None,
                     proj = None):
    driver = gdal.GetDriverByName(drivername)
    ds = driver.Create(filename, cols, rows, bands, datatype)
    if ds is None:
        raise ValueError("dataset has 'None' value")
    if geot is not None and ds is not None:
        ds.SetGeoTransform(geot)
    if proj is not None and ds is not None:
        ds.SetProjection(proj)
    return ds


def createMask(rasterPath, minValue, maxValue, band=1):
    """
    Create a mask from a raster band

    Args:
        rasterPath: path to raster
        minValue: minimum data value
        maxValue: maximum data value
        band: raster band to use (default: 1)

    Returns:
        Integer array with a value of 0 where the input raster band is < minValue or > maxValue

    """
    array = getRasterBandAsArray(rasterPath, band)
    return np.where((array < minValue) | (array > maxValue), 0, 1)

def getBoundingBox(rasterPath):
    rows, cols, geot = getGeoTransformAndSize(rasterPath)
    if not any(i is None for i in [rows, cols, geot]):
        return getBoundingBox_geot(geot, rows, cols)
    else:
        return None

def getBoundingBox_geot(geot, nrow, ncol):
    """
    Returns bounding box coordinates as a tuple of (minX, minY, maxX, maxY)
    Args:
        geot: gdal affine geotransform for rater
        nrow: number of rows
        ncol: number of columns

    Returns:
        bounding box coordinates as a tuple of (minX, minY, maxX, maxY)

    """
    r = geot[0] + ncol * geot[1]
    b = geot[3] + nrow * geot[5]
    return (geot[0], b, r, geot[3])

def getCellAddressOfPoint(x, y, geot):
    """

    Args:
        x:
        y:
        geot:

    Returns:

    """
    col = math.floor((x - geot[0]) / geot[1])
    row = math.floor((geot[3] - y) / abs(geot[5]))
    return (row, col)

def getGeoTransform(rasterPath):
    """
    Get the affine geotransformation information for a raster dataset

    Args:
        rasterPath: path to rater dataset

    Returns: 6 element list if successful, None if not successful

    """
    ds = gdal.Open(rasterPath)
    if ds is not None:
        geot = ds.GetGeoTransform()
        ds = None
        return geot
    else:
        return None

def getGeoTransformAndSize(rasterPath):
    """
    Get affine transformation information and the number of rows and columns for a raster dataset.

    Args:
        rasterPath: Path to raster dataset.

    Returns:
        int, int, list: number of rows, number of columns, geotransform (six item list)

    """
    geot = getGeoTransform(rasterPath)
    if geot is not None:
        ds = gdal.Open(rasterPath)
        if ds is not None:
            rows = ds.RasterYSize
            cols = ds.RasterXSize
            return rows, cols, geot
        else:
            return None, None, geot
    else:
        return None, None, None

def getOffsetGeot(row, col, geot):
    newgeot = [
        geot[0] + (col * geot[1]),
        geot[1],
        0.0,
        geot[3] + (row * geot[5]),
        0.0,
        geot[5]
    ]
    return newgeot

def getProjection(rasterPath):
    ds = openGDALRaster(rasterPath)
    return ds.GetProjection()

def getRasterAsArray(rasterPath):
    """
    Returns raster as numpy array

    Args:
        rasterPath: Path to input rater

    Returns:
        numpy array if file exists or None if it does not

    """
    if os.path.isfile(rasterPath):
        ds = gdal.Open(rasterPath)
        array = ds.ReadAsArray()
        ds = None
        return array
    else:
        return None

def getRasterBandAsArray(rasterPath, band = 1):
    """
    Returns raster band as 2d numpy array

    Args:
        rasterPath: Path to input raster

    Returns:
        numpy array if band exists or None if band does not exist

    """
    if os.path.isfile(rasterPath):
        ds = gdal.Open(rasterPath)
        bands = ds.RasterCount
        if band > 0 and band <= bands:
            array = ds.GetRasterBand(band).ReadAsArray()
            ds = None
            return array
        else:
            return None
    else:
        return None

def getXYResolution(rasterPath):
    """
    Get X and Y pixel resolution from a rater dataset

    Args:
        rasterPath: path to raster dataset

    Returns:
        X resolution (positive), Y resolution (negative) on success or None on failure

    """
    geot = getGeoTransform(rasterPath)
    if geot is not None: return geot[1], geot[5]
    else: return None

def greaterThan(valueRaster, compareValue, outputraster, valuetrue=1, valuefalse=0):
    """

    Args:
        valueRaster:
        compareRaster:
        outputraster:
        valuetrue:
        valuefalse:

    Returns:

    """

    values = getRasterBandAsArray(valueRaster, 1)
    compare = getRasterBandAsArray(compareValue, 1)
    result = np.where(values > compare, valuetrue, valuefalse)
    resultMasked = maskArray(result, values, -9999)
    writeArrayAsRaster(outputraster, resultMasked, result.shape[0], result.shape[1], getGeoTransform(valueRaster),
                       getProjection(valueRaster), nodata=-9999)

    return None

def lessThan(raster1, raster2, outputraster, valuetrue, valuefalse):
    """

    Args:
        raster1:
        raster2:
        outputraster:
        valuetrue:
        valuefalse:

    Returns:

    """
    return None

def linearIndexOfCoordinates(x, y, geot, rows, cols, band=1):
    row, col = addressOfCoordinates(x, y, geot)
    idx = np.add(np.add(np.multiply(row, cols), col), rows*cols*(band-1))
    return idx.astype(int)

def linearTakeBand(values, band_indices):
    """
    Get 2d array of band values from a multiband raster of shape (bands, rows, columns)
    Args:
        values: input 3d array containing data values
        band_indices: 2d array containing indices to bands in the value raster

    Returns:
        2d array where the value in the array corresponds the band value from band_indices

    """
    _, nR, nC = values.shape
    idx = nC*nR*band_indices + nC*np.arange(nR)[:, None] + np.arange(nC) #convert 2d indices to linear indices
    return np.take(values, idx), idx

def maskArray(array, mask, nodata=-9999):
    """
    Replace array values with a no data value where a mask is false (0)

    Args:
        array: input array to be masked
        mask: boolean array or integer array with values of 0 (false) and 1 (true)
        nodata: value to write where mask is false (default: -9999)

    Returns:
        array with nodata where mask is false

    """
    if array.shape != mask.shape:
        print("error masking array: array shapes are different", array.shape, mask.shape)
    return np.where(mask, array, nodata)

def maskRaster(rasterPath, array, nodata=-9999, band=1):
    """
    Replace values in a raster band with no data where another array is equal to no data

    Args:
        rasterPath: Path of raster to mask.
        array: Array to mask with (must be same shape as array from raster).
        nodata: No data value of array (default: -9999).
        band: Band of raster to mask (default: 1).

    Returns:

    """
    mask = np.where(array==nodata, 0, 1)
    ds = openGDALRaster(rasterPath, gdal.GA_Update)
    band = ds.GetRasterBand(band).ReadAsArray()
    ds.GetRasterBand(1).WriteArray(maskArray(band, mask, nodata))
    ds.GetRasterBand.SetNoDataValue(nodata)
    ds = None
    return None

def maskRasterWithValues(inputraster, minvalue, maxvalue, method="include", band=1):
    ds = openGDALRaster(inputraster, gdal.GA_Update)
    banddata = ds.GetRasterBand(band).ReadAsArray()
    nodata = ds.GetRasterBand(band).GetNoDataValue()
    writedata = None
    if method == "include":
        writedata = np.where((banddata >= minvalue) & (banddata <= maxvalue), banddata, nodata)
    elif method == "exclude":
        writedata = np.where((banddata < minvalue) & (banddata > maxvalue), banddata, nodata)

    if writedata is not None:
        ds.GetRasterBand(band).WriteArray(writedata)
    ds=None

def maskRasterWithRaster(inputraster, maskraster, inputband=1, maskband=1):
    """
    Mask raster with another raster according to no data extent.

    Args:
        inputraster: Path of raster to be masked.
        maskraster: Path of raster to use as mask.
        inputband: Band of input raster to be masked (default: 1).
        maskband: Band of mask raster to use as mask (default: 1).

    Returns:

    """
    ds = openGDALRaster(inputraster, gdal.GA_Update)
    dsmask = openGDALRaster(maskraster)
    band = ds.GetRasterBand(inputband).ReadAsArray()
    maskarray = dsmask.GetRasterBand(maskband).ReadAsArray()
    nodata = dsmask.GetRasterBand(maskband).GetNoDataValue()
    print(nodata)
    ds.GetRasterBand(inputband).WriteArray(maskArray(band, np.where(maskarray==nodata, 0, 1), nodata))
    ds.GetRasterBand(inputband).SetNoDataValue(nodata)
    ds = None
    dsmask = None

def openGDALRaster(rasterPath, access=gdal.GA_ReadOnly):
    ds = gdal.Open(rasterPath, access)
    if ds is not None:
        return ds

def percentileForAllBands(multipath, outpath, maskpath=None):
    multi = getRasterAsArray(multipath)
    outmulti = np.empty(multi.shape)
    mean = np.mean(multi, axis=0)
    sd = np.std(multi, axis=0)
    for band in range(multi.shape[0]):
        data = multi[band, :, :]
        outmulti[band, :, :] = stats.norm.cdf(data, loc=mean, scale=sd)*100
    writeArrayAsRaster(outpath, outmulti, outmulti.shape[1], outmulti.shape[2], getGeoTransform(multipath), getProjection(multipath))
    if maskpath is not None:
        for i in range(outmulti.shape[0]):
            maskRasterWithRaster(outpath, maskpath, i+1, 1)

def percentileMultiband(multi, index):
    """
    Calculate the percentile of a specified value at a position in a multiband raster.

    Args:
        multi: Multiband numpy array.
        index: Numpy array containing the band index.

    Returns:
        2d array result (percentile of value based on all bands), and score (value of requested band).

    """
    mean = np.mean(multi, axis=0) #mean of all bands at each row,col
    sd = np.std(multi, axis=0) #standard deviation of all bands at each row,col
    score, idx = linearTakeBand(multi, index) #value of index band at each row,col
    result = stats.norm.cdf(score, loc=mean, scale=sd)*100 #percentile
    return result, score

def percentileOfMultibandIndex(datapath, index, percentilepath, scorepath=None, mask = None):
    """
    For a multiband raster, calculates the percentile of a band value at each row,col

    Args:
        datapath: input path to multiband raster
        index: 2d array with each row,col containing an index to a band in data path
        percentilepath: output path for 2d raster of percentile values
        scorepath: ouput path for 2d raster of the scores of each band index (optional)
        maskpath: boolean array to mask outputs (optional)

    Returns:
        None

    """
    multi = getRasterAsArray(datapath)
    if index is not None and index.shape == multi.shape[1:]:
        result, score = percentileMultiband(multi, index)
        if mask is None or mask.shape != result.shape:
            mask = np.ones((result.shape))

        ds = gdal.Open(datapath)
        writeArrayAsRaster(percentilepath, maskArray(result, mask), rows=result.shape[0], cols=result.shape[1],
                          geot=ds.GetGeoTransform(), srs=ds.GetProjection())
        if scorepath is not None:
            writeArrayAsRaster(scorepath, maskArray(score, mask), rows=result.shape[0], cols=result.shape[1],
                              geot=ds.GetGeoTransform(), srs=ds.GetProjection())
        ds = None
    else:
        print("problem with input index array", index.shape, multi.shape[1:])

    return None

def remapValues(inpath, outpath, remap_values, new_values, nodata = -9999.0, band = 1, proj=None):
    """
    Remap values in a raster, writes a new raster
    Args:
        inpath: path to input raster
        outpath: path to write output raster
        replace_values: values to replace
        new_values: values to write, must correspond to replace_values
        nodata: no data value (default: -9999)
        band: raster band to use (default: 1)
        proj: projection for new raster

    Returns:

    """
    ds = openGDALRaster(inpath)
    values = ds.GetRasterBand(band).ReadAsArray()
    result = replaceValues(values, remap_values, new_values, nodata)
    if proj is None:
        proj = ds.GetProjection()
    writeArrayAsRaster(outpath, result, ds.RasterYSize, ds.RasterXSize, ds.GetGeoTransform(), proj, nodata=nodata)

def replaceValues(array, remap_values, new_values, nodata=-9999.0):
    """
    Replaces specified values in an array with new values. nodata values will remain nodata values.
    Args:
        array: input array
        remap_values: values to replace/remap
        new_values: values to replace remap_values
        nodata: nodata value

    Returns:
        remapped numpy array with same dimensions of `array`

    """
    flat = array.flatten()  # flatten to 1d array
    isort = np.argsort(remap_values)  # get sorted indices
    i = np.searchsorted(remap_values[isort], flat)   # get indices where remap values occur in original raster
    output = new_values[isort][i]  # replace with new values
    output = np.where(flat == nodata, nodata, output)
    output = np.reshape(output, array.shape)
    return output

def writeArrayAsRaster(path, array, rows, cols, geot, srs, nodata=-9999, nan=-9999, datatype=gdal.GDT_Float32, drivername = 'GTiff'):
    """
    Write array to a raster dataset

    Args:
        path: output file for raster
        array: array containing data
        rows: number of rows in array
        cols: number of columns in array
        geot: affine geotransformation for the output raster
        srs: spatial reference for the output raster (default: None)
        nodata: no data value for the output raster (default: -9999)
        nan: value in array that should be written as nodata (default: -9999)
        datatype: gdal data type of output raster (default: GDT_Float32)
        drivername: Name of GDAL driver to use to create raster (default: 'GTiff')

    Returns:
        None

    """
    driver = gdal.GetDriverByName(drivername)
    bands = 1
    multi = False
    if len(array.shape) == 3:
        bands = array.shape[0]
        multi = True
    ds = driver.Create(path, xsize=cols, ysize=rows, bands=bands, eType=datatype)
    ds.SetProjection(srs)
    ds.SetGeoTransform(geot)
    array = np.where((array==np.nan) | (array==nan), nodata, array)
    for band in range(bands):
        print('writing band', band+1, 'of', bands)
        if multi:
            ds.GetRasterBand(band+1).WriteArray(array[band,:,:])
        else:
            ds.GetRasterBand(band + 1).WriteArray(array)
        ds.GetRasterBand(band+1).SetNoDataValue(nodata)
        ds.GetRasterBand(band+1).FlushCache()
    ds = None
    return None


from osgeo import gdal, ogr, osr
import os
import pysal as ps
import pandas as pd

def vector_test():
    return "this is the vector module of the gispy package"

def copyFields(lyr, lyrdefn):
    """
    Create fields in a new layer based on fields from an existing OGRLayerDefn.

    Args:
        lyr: OGRLayer to create fields for.
        lyrdefn: OGRLayerDefn to copy from.

    Returns:
        OGRLayer with fields from lyrdefn added.

    """
    for i in range(0, lyrdefn.GetFieldCount()):
        fldefn = lyrdefn.GetFieldDefn(i)
        lyr.CreateField(fldefn)
    return lyr

def copyFieldValues(newfeat, copyfeat, defn):
    """
    Copy field values from one feature to another.

    Args:
        newfeat: OGRFeature to copy to
        copyfeat: OGRFeature containing field values to copy
        defn: OGRFieldDefinition of copyfeat

    Returns:

    """
    for i in range(0, defn.GetFieldCount()):
        newfeat.SetField(defn.GetFieldDefn(i).GetNameRef(), copyfeat.GetField(i))

def createFields(lyr, fieldnames, fieldtype = ogr.OFTReal):
    """
    Add fields to a shapefile layer

    Args:
        lyr: ogr layer to add fields to
        fieldnames: names of fields to add
        fieldtype: data type of fields to add

    Returns:
        ogr.Layer: input layer with fields added

    """
    for name in fieldnames:
        field = ogr.FieldDefn(name, fieldtype)
        index = lyr.FindFieldIndex(name, 1)
        if index is not -1:
            lyr.DeleteField(index)
        lyr.CreateField(field)
    return lyr

def createIDField(inshp, fname = "ID"):
    ds = ogr.Open(inshp, 1)
    lyr = ds.GetLayer(0)
    lyr = createFields(lyr, [fname], ogr.OFTInteger)
    feat = lyr.GetNextFeature()
    while feat:
        fid = feat.GetFID()
        feat.SetField(fname, fid)
        lyr.SetFeature(feat)
        feat = lyr.GetNextFeature()
        if fid % 50000 == 0:
            print(fid)
    ds.Destroy()
    return None

def createOGRDataSource(filename, driver='ESRI Shapefile'):
    """
    Create OGRDataSource
    Args:
        filename: Path of file to create
        driver: Name of OGR Driver to use (default: 'ESRI Shapefile')

    Returns:
        OGRDataSource

    """
    driver = ogr.GetDriverByName(driver)
    if os.path.exists(filename):
        driver.DeleteDataSource(filename)
    ds = driver.CreateDataSource(filename)
    return ds

def dbf2DF(dbfile):
    """
    Converts DBF file to pandas data frame. From: https://gist.github.com/mhweber/a2d5863fcc4d52376481

    Args:
        dbfile: path to DBF file

    Returns:
        pandas data frame

    """
    dbf = ps.lib.io.open(dbfile)
    dbf_dict = {col: dbf.by_col(col) for col in dbf.header} # convert dbf to dictionary
    df = pd.DataFrame(dbf_dict)
    return df

def deleteFields(filename, fields):
    """
    Delete fields from shapefile

    Args:
        filename: Path to shapefile.
        fields: List of fields to delete

    Returns:

    """
    ds = gdal.OpenEx(filename, gdal.OF_VECTOR | gdal.OF_UPDATE)
    layername = getFilenameWithoutExtenstion(filename)
    for field in fields:
        sql = 'ALTER TABLE ' + layername + ' DROP COLUMN ' + field
        ds.ExecuteSQL(sql)
    ds = None

def getFieldValues(dataset, fieldName):
    """
    Get list of all values for shapefile field

    Args:
        dataset: path to shapefile
        fieldName: name of field

    Returns:
        list of field values and FIDs if successful, None if not successful

    """
    values = []
    fids = []
    ds = ogr.Open(dataset)
    lyr = ds.GetLayer()
    if lyr is not None:
        for feat in lyr:
            values.append(feat.GetField(fieldName))
            fids.append(feat.GetFID())
        return values, fids
    else:
        return None

def getFilenameWithoutExtenstion(filepath):
    """
    Return the name of a file without the director or extension

    Args:
        filepath: Path to file.

    Returns:
        string: name of file without extension

    """
    base = os.path.basename(filepath)
    return os.path.splitext(base)[0]

def joinZonalStatsToSHP(inshp, zsresult, id, stats, fieldnames, stattype=ogr.OFTReal):
    """
    Join zonal stats results to shapefile

    Args:
        inshp: shapefile to add zonal stats to
        zsresult: result of zonal stats (from rasterstats package)
        id: name of zonal stat with FID of feature corresponding to stats
        stats: names of statistics to join
        fieldnames: names of fields to create (corresponding to stats)
        stattype: ogr OFT data type of field to create (default: ogr.OFTReal)

    Returns:
        None

    """
    print("starting join")
    ds = ogr.Open(inshp, 1)
    lyr = ds.GetLayer(0)
    lyr = createFields(lyr, fieldnames, stattype)
    print("fields created", fieldnames)
    # layerDefinition = lyr.GetLayerDefn()
    # for i in range(layerDefinition.GetFieldCount()):
    #     print layerDefinition.GetFieldDefn(i).GetName()
    iter = 0
    for result in zsresult:
        # print result
        # print(int(result[id]))
        feat = lyr.GetFeature(int(result[id]))
        for i in range(len(stats)):
            value = result[stats[i]]
            if stattype == ogr.OFTReal:
                if result[stats[i]] == None:
                    value = -9999.0
                else:
                    value = result[stats[i]]*1.0
            if feat:
                # if feat.GetFID()==62:
                #     print fieldnames[i], value, result[id]
                if value != -9999.0:
                    feat.SetField(fieldnames[i], value)
        iter += 1
        if (iter % 100000 == 0):
            print("joined", iter, "of", lyr.GetFeatureCount())
        if feat and value != -9999.0:
            lyr.SetFeature(feat)
    # else:
    #     feat = lyr.GetNextFeature()
    #     while feat:
    #         featid = feat.GetField(id)
    #         result = next((item for item in zsresult if item[id] == int(featid)), None)
    #         #print result
    #         for i in range(len(stats)):
    #             if stattype == ogr.OFTReal:
    #                 if result == None or result[stats[i]] == None:
    #                     value = -9999.0
    #                 else:
    #                     value = result[stats[i]]*1.0
    #             feat.SetField(fieldnames[i], value)
    #             #print fieldnames[i], value, inshp
    #         iter += 1
    #         if (iter % 100000 == 0):
    #             print "joined", iter, "of", lyr.GetFeatureCount()
    #         lyr.SetFeature(feat)
    #         feat = lyr.GetNextFeature()
    ds.Destroy()
    return None

def joinZonalStatsToSHP_rasterstats(inshp, zsresult, id, stats, fieldnames, stattype=ogr.OFTReal):
    """
    Join zonal stats results to shapefile

    Args:
        inshp: shapefile to add zonal stats to
        zsresult: result of zonal stats (from rasterstats package)
        id: FID of feature corresponding to stats
        stats: names of statistics to join
        fieldnames: names of fields to create (corresponding to stats)
        stattype: ogr OFT data type of field to create (default: ogr.OFTReal)

    Returns:
        None

    """
    ds = ogr.Open(inshp, 1)
    lyr = ds.GetLayer(0)
    lyr = createFields(lyr, fieldnames, stattype)
    for result in zsresult:
        feat = lyr.GetFeature(int(result[id]))
        for i in range(len(stats)):
            feat.SetField(fieldnames[i], result["properties"][stats[i]])
        lyr.SetFeature(feat)
    ds.Destroy()
    return None

def fieldExists(lyr, fieldname):
    """
    Determine if field exists in a layer.

    Args:
        lyr: OGR Layer.
        fieldname: Name of field.

    Returns:
        bool: True if field exists, False if it does not.

    """
    defn = lyr.GetLayerDefn()
    if defn.GetFieldIndex(fieldname) is not -1:
        return True
    else:
        return False

def openOGRDataSource(file, access=0):
    """
    Open and existing OGRDataSource.

    Args:
        file: Path of file to open.
        access: Access type, 0 read only, 1 update (default: 0)

    Returns:
        OGRDataSource

    """
    ds = ogr.Open(file, access)
    if ds is not None:
        return ds
    else:
        print('Problem opening shapefile')

def reprojectShapefileLayer(shapefile, newshapefile, newsrs):
    """
    Create a new shapefile with a different projection.

    Args:
        shapefile: Path of shapefile to reproject.
        newshapefile: Path of shapefile to create with new projection.
        newsrs: OSRSpatialReference of the new shapefile.

    Returns:

    """
    ds = openOGRDataSource(shapefile)
    layer = ds.GetLayer()
    srs = layer.GetSpatialRef()
    transform = osr.CoordinateTransformation(srs, newsrs)

    outds = createOGRDataSource(newshapefile)
    outlyr = outds.CreateLayer(getFilenameWithoutExtenstion(newshapefile), srs=newsrs, geom_type=layer.GetGeomType())
    copyFields(outlyr, layer.GetLayerDefn())
    defn = outlyr.GetLayerDefn()
    feat = layer.GetNextFeature()
    while feat:
        geom = feat.GetGeometryRef()
        geom.Transform(transform)
        outfeat = ogr.Feature(defn)
        outfeat.SetGeometry(geom)
        copyFieldValues(outfeat, feat, defn)
        outlyr.CreateFeature(outfeat)
        outfeat = None
        feat = layer.GetNextFeature()
    ds = None
    outds = None

def saveFilteredLayerAsShapefile(filename, layer, filter, driver='ESRI Shapefile', close=False):
    """
    Filter OGRLayer with SQL statement and save as a shapefile
    Args:
        filename: Path to save new shapefile.
        layer: Layer to apply SQL filter to.
        filter: Filter to apply to layer.
        driver: Name of OGR driver used to create the new shapefile (default: ESRI Shapefile)
        close: If the new OGRDataSource and OGRLayer should be closed. False: data source and layer returned, True: data source and layer closed (default: False)

    Returns:
        OGRDataSource and OGRLayer if 'close' is False.

    """
    layer.SetAttributeFilter(filter)
    return saveLayerAsShapefile(filename, layer, driver, close)

def saveLayerAsShapefile(filename, layer, driver='ESRI Shapefile', close=False):
    """
    Save a vector layer to a new OGR DataSource.

    Args:
        filename: Name of new data source/shapefile
        layer: Layer to copy.
        driver: Name of driver to create DataSource (default: ESRI Shapefile)
        close: Close created DataSource and layer, if True new ds and new layer will be deleted, if False they will be returned (default: False)

    Returns:
        OGRDataSource and OGRLayer, if 'close' is False.

    """
    ds = createOGRDataSource(filename, driver)
    newlyr = ds.CopyLayer(layer, getFilenameWithoutExtenstion(filename))
    if close:
        del ds, newlyr
        return None
    else:
        return ds, newlyr



from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


setup(name='gispy',
      version='0.0.10',
      description='GIS functions with GDAL/OGR',
      long_description=readme(),
      classifiers=[
            'Development Status :: 3 - Alpha',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Natural Language :: English',
            'Topic :: Scientific/Engineering :: GIS',
      ],
      keywords='GIS, raster, vector',
      url='https://github.com/konradhafen/gispy',
      author='Konrad Hafen',
      author_email='khafen74@gmail.com',
      license='GPLv3',
      packages=['gispy'],
      install_requires=[
            'gdal',
            'numpy',
            'scipy',
      ],
      include_package_data=True,
      zip_safe=False)



from osgeo import gdal, ogr, osr
import numpy as np
from scipy import stats
import math
import struct
import os
import gispy.raster as raster
import gispy.vector as vector


def bboxToOffsets(bbox, geot):
    col1 = int((bbox[0] - geot[0]) / geot[1])
    col2 = int((bbox[1] - geot[0]) / geot[1]) + 1
    row1 = int((bbox[3] - geot[3]) / geot[5])
    row2 = int((bbox[2] - geot[3]) / geot[5]) + 1
    return [row1, row2, col1, col2]

def clipByFeature(inputdir, outputdir, rasterfiles, shapefile, fieldname, nodata=-9999, xres=None, yres=None):
    """
    Clip multiple rasters by features in a shapefile. Creates a new directory for each feature and saves clipped rasters in the directory

    Note:
        All input rasters and shapefiles should have the same spatial reference

    Args:
        inputdir: directory where rasters to be clipped are located
        outputdir: directory where a new directory for each clipped raster will be created
        rasterfiles: list of raster file names and extensions to be clipped. e.g. ['raster1.tif', 'raster2.tif]
        shapefile: shapefile containing features to clip with
        fieldname: name of field to select feature with
        nodata: nodata value (default: -9999)
        xres: x resolution of output raster (default: None), with default resolution is taken from the input rater
        yres: y resolution of output rater (defaul: None), with default resolution is taken from the input rater

    Returns:
        None

    """
    fieldValues, fids = vector.getFieldValues(shapefile, fieldname)
    fieldValues_unique = list(set(fieldValues)) #get unique field values (otherwise the same operation may be done twice)
    for value in fieldValues_unique:
        dirvalue = outputdir + "/" + str(value)
        if not os.path.isdir(dirvalue):
            os.mkdir(dirvalue)
        for raster in rasterfiles:
            infile = inputdir + "/" + str(raster)
            if os.path.exists(infile):
                clipRasterWithPolygon(infile, shapefile, dirvalue + "/" + str(raster), nodata=nodata, xres=xres, yres=yres,
                                      fieldValue=value, field=fieldname)
    return None


def clipRasterWithPolygon(rasterpath, polygonpath, outputpath, nodata=-9999, dstSrs=None, xres=None, yres=None,
                          field=None, fieldValue=None):
    """

    Args:
        rasterpath: raster to clip
        polygonpath: shapefile containing features to clip with
        outputpath: path of output clipped raster
        nodata: nodata value (default: -9999)
        dstSrs: spatial reference for output raster if different than input raster (default: None)
        xres: x resolution of output raster (default: None, resolution of input raster)
        yres: y resolution of output raster (default: None, resolution of input raster)
        field: name of shapefile field to select feature for clipping
        fieldValue: field values to specify the clipping feature

    Returns:

    """
    if xres is None or yres is None: xres, yres = raster.getXYResolution(rasterpath)
    if field is not None and fieldValue is not None: wexp = str(field) + " = \'" + str(fieldValue) + "\'"
    else: wexp = None

    warpOptions = gdal.WarpOptions(format='GTiff', cutlineDSName=polygonpath, cropToCutline=True, cutlineWhere=wexp,
                                   xRes=xres, yRes=abs(yres), dstNodata=nodata)
    gdal.WarpOptions()
    gdal.Warp(outputpath, rasterpath, options=warpOptions)
    return None

def polygonToRaster(rasterpath, vectorpath, fieldname, rows, cols, geot, prj=None, drivername='GTiff', allcells=False,
                    nodata=-9999, datatype = gdal.GDT_Float32, islayer=False):
    """
    Convert polygon shapefile to raster dataset.

    Args:
        rasterpath: Path of raster to be created.
        vectorpath: Path of polygon shapefile to rasterize.
        fieldname: Name of shapefile field to use as values in new raster.
        rows: Number of rows in new raster.
        cols: Number of columns in new raster.
        geot: Affine geotransform of new raster.
        prj: Spatial reference of new raster.
        drivername: Name of GDAL driver to use to create raster (default: 'GTiff')
        allcells: If all cells intersected by polygons should be rasterized, or just when polygon includes cell center (defaul: False).
        nodata: No data value.
        datatype: GDAL datatype of new raster (default: gdal.GDT_Float32).
        islayer (bool): True if vector path is an OGRLayer (default: False)

    Returns:

    """
    if islayer:
        lyr = vectorpath
    else:
        inds = ogr.Open(vectorpath)
        lyr = inds.GetLayer()
    driver = gdal.GetDriverByName(drivername)
    outds = driver.Create(rasterpath, cols, rows, 1, datatype)
    outds.SetProjection(prj)
    outds.SetGeoTransform(geot)
    band = outds.GetRasterBand(1).ReadAsArray()
    band.fill(nodata)
    outds.GetRasterBand(1).WriteArray(band)
    outds.GetRasterBand(1).SetNoDataValue(nodata)

    ALL_TOUCHED = 'FALSE'
    if allcells: ALL_TOUCHED = 'TRUE'

    if vector.fieldExists(lyr, fieldname):
        status = gdal.RasterizeLayer(outds, [1], lyr, options=['ALL_TOUCHED='+ALL_TOUCHED, 'ATTRIBUTE='+fieldname, 'NODATA='+str(nodata)])
        if status is not 0:
            print("Rasterize not successful")
    else:
        print("Rasterize field does not exist")

    if inds: inds = None
    outds = None
    return None

def rasterValueAtPoints(pointshapefile, rasterpath, fieldname, datatype=ogr.OFTReal, idxfield=None):
    """
    Get the value of a raster at point locations.

    Args:
        pointshapefile: Path to point shapefile.
        rasterpath: Path to raster dataset.
        fieldname: Name of field to create and write raster values.
        datatype: OGR datatype of created field (default: ogr.OFTReal)
        idxfield: Shapefile field containing band number to use on multipband rasters (default: None, get value from band 1)

    Returns:

    """
    ras = raster.openGDALRaster(rasterpath)
    geot = ras.GetGeoTransform()
    shp = vector.openOGRDataSource(pointshapefile, 1)
    lyr = shp.GetLayer()
    # if lyr.GetGeomType() is not ogr.wkbPoint:
    #     print "incorrect geometry type, should be points", lyr.GetGeomType()
    #     return None
    vector.createFields(lyr, [fieldname], datatype)
    feat = lyr.GetNextFeature()
    while feat:
        nband = 1
        if idxfield is not None:
            nband = feat.GetField(idxfield)
        geom = feat.GetGeometryRef()
        row, col = raster.getCellAddressOfPoint(geom.GetX(), geom.GetY(), geot)
        if nband <= 0 or nband > ras.RasterCount:
            value = -9999.0
        else:
            value = struct.unpack('f'*1, ras.GetRasterBand(int(nband)).ReadRaster(xoff=col, yoff=row, xsize=1, ysize=1,
                                                                      buf_xsize=1, buf_ysize=1,
                                                                      buf_type=gdal.GDT_Float32))[0]
        feat.SetField(fieldname, value)
        lyr.SetFeature(feat)
        feat = lyr.GetNextFeature()
    lyr = None
    shp.Destroy()

def rasterZonesFromVector_delta(vectorpath, rasterpath, outputpath, deltavalue=10000.0, deltatype='percent', minvalue = 0.0, outdiff=None):
    print(deltavalue, minvalue)
    np.set_printoptions(suppress=True)
    rasterds = raster.openGDALRaster(rasterpath)
    vectords = vector.openOGRDataSource(vectorpath)
    print("data opened")
    lyr = vectords.GetLayer()
    geot = rasterds.GetGeoTransform()
    nodata = rasterds.GetRasterBand(1).GetNoDataValue()
    outds = raster.createGDALRaster(outputpath, rasterds.RasterYSize, rasterds.RasterXSize, geot=geot,
                                    datatype=gdal.GDT_Int32)
    if outdiff is not None:
        diffds = raster.createGDALRaster(outdiff, rasterds.RasterYSize, rasterds.RasterXSize, geot=geot,
                                        datatype=gdal.GDT_Float32)
        diffds.GetRasterBand(1).SetNoDataValue(-9999.0)
        diffds.SetProjection(rasterds.GetProjection())
    print("datacreated")
    #outds.GetRasterBand(1).Fill(-1)
    print("filled")
    outds.GetRasterBand(1).SetNoDataValue(-1)
    outds.SetProjection(rasterds.GetProjection())
    outofbounds = []
    feat = lyr.GetNextFeature()
    iter = 0
    print("starting loop")
    while feat:
        id = feat.GetFID()
        tmpds = vector.createOGRDataSource('temp', 'Memory')
        tmplyr = tmpds.CreateLayer('polygons', None, ogr.wkbPolygon)
        tmplyr.CreateFeature(feat.Clone())
        offsets = bboxToOffsets(feat.GetGeometryRef().GetEnvelope(), geot)
        for i in range(0, len(offsets)):
            if offsets[i] < 0:
                offsets[i] = 0
        if not any(x < 0 for x in offsets):
            array = rasterds.GetRasterBand(1).ReadAsArray(offsets[2], offsets[0], (offsets[3] - offsets[2]),
                                                          (offsets[1] - offsets[0]))
            outarray = outds.GetRasterBand(1).ReadAsArray(offsets[2], offsets[0], (offsets[3] - offsets[2]),
                                                          (offsets[1] - offsets[0]))
            newgeot = raster.getOffsetGeot(offsets[0], offsets[2], geot)
            tmpras = raster.createGDALRaster('', offsets[1] - offsets[0], offsets[3] - offsets[2],
                                             datatype=gdal.GDT_Byte,
                                             drivername='MEM', geot=newgeot)
            gdal.RasterizeLayer(tmpras, [1], tmplyr, burn_values=[1])
            tmparray = tmpras.ReadAsArray()
            featarray = np.ma.MaskedArray(array,
                                              mask=np.logical_or(array < minvalue, np.logical_not(tmparray)))
            featmean = np.ma.mean(featarray)
            #print id

            if featmean != nodata:
                median = np.ma.median(featarray)
                #diff = (abs(featarray - median) / median) * 100.0
                diff = np.divide((featarray - median), array)
                print("diff", np.min(diff))
                # diff = (abs(featarray - featmean) / featmean) * 100.0

                maskarray = np.ma.MaskedArray(array,
                                              mask=np.logical_or(np.ma.getmask(featarray),
                                                                 np.logical_or(diff > deltavalue, array < minvalue)))
                print("mask", np.min(featarray), np.min(array[array >= 125]))
                maskarray.set_fill_value(-1)
                maskarray = maskarray.filled()
                maskarray = np.where(maskarray>=0, id, np.where(outarray>=0, outarray, maskarray))
                # if id == 1954:
                #     print "median", median, "mean", featmean
                #     print "feat array"
                #     print featarray
                #     print "tmp array"
                #     print tmparray
                #     print "array"
                #     print np.around(array)
                #     print "diff"
                #     print np.around(diff)
                #     print "result"
                #     print maskarray
                outds.GetRasterBand(1).WriteArray(maskarray, offsets[2], offsets[0])
                if outdiff is not None:
                    diff.set_fill_value(-9999.0)
                    diff = diff.filled()
                    diff = np.where(diff >= -9998.9, diff, -9999.0)
                    diffds.GetRasterBand(1).WriteArray(diff, offsets[2], offsets[0])

        else:
            print("out of bounds", feat.GetFID())
            outofbounds.append(feat.GetFID())
        tmpras = None
        tmpds = None
        iter += 1
        if (iter % 10000 == 0):
            print("iter", iter, "of", lyr.GetFeatureCount())
        feat = lyr.GetNextFeature()
    #outds = None
    print("done")
    return None

def setFeatureStats(fid, min=None, max=None, sd=None, mean=None, median=None, sum=None, count=None, majority=None,
                    deltamed=None, names=['min', 'mean', 'median', 'max', 'sd', 'sum', 'count', 'majority', 'fid', 'deltamed']):
    featstats = {
        names[0]: min,
        names[1]: mean,
        names[2]: median,
        names[3]: max,
        names[4]: sd,
        names[5]: sum,
        names[6]: count,
        names[7]: majority,
        names[8]: fid,
        names[9]: deltamed
    }
    return featstats

def zonalStatistics(vectorpath, rasterpath, idxfield="fid", snodata=-9999.0,
                    names=['min', 'mean', 'median', 'max', 'sd', 'sum', 'count', 'majority', 'fid', 'deltamed']):
    names[8] = idxfield
    rasterds = raster.openGDALRaster(rasterpath)
    vectords = vector.openOGRDataSource(vectorpath)
    lyr = vectords.GetLayer()
    geot = rasterds.GetGeoTransform()
    nodata = rasterds.GetRasterBand(1).GetNoDataValue()
    zstats=[]
    feat = lyr.GetNextFeature()
    iter = 0
    while feat:
        if feat.GetGeometryRef() is not None:
            tmpds = vector.createOGRDataSource('temp', 'Memory')
            tmplyr = tmpds.CreateLayer('polygons', None, ogr.wkbPolygon)
            tmplyr.CreateFeature(feat.Clone())
            offsets = bboxToOffsets(feat.GetGeometryRef().GetEnvelope(), geot)
            newgeot= raster.getOffsetGeot(offsets[0], offsets[2], geot)
            tmpras = raster.createGDALRaster('', offsets[1]-offsets[0], offsets[3]-offsets[2], datatype=gdal.GDT_Byte, drivername='MEM', geot=newgeot)
            gdal.RasterizeLayer(tmpras, [1], tmplyr, burn_values=[1])
            tmparray = tmpras.ReadAsArray()
            array = rasterds.GetRasterBand(1).ReadAsArray(offsets[2], offsets[0], (offsets[3] - offsets[2]),
                                                          (offsets[1] - offsets[0]))
            if idxfield == "FID" or idxfield == "fid":
                id = feat.GetFID()
            else:
                id = feat.GetField(idxfield)

            if array is not None and np.logical_or(array == nodata, np.logical_not(tmparray)) is not None:
                maskarray = np.ma.MaskedArray(array, mask=np.logical_or(array==nodata, np.logical_not(tmparray)))

                zstats.append(setFeatureStats(id, min=float(maskarray.min()), mean=float(maskarray.mean()),
                                              max=float(maskarray.max()), sum=float(maskarray.sum()), sd=float(maskarray.std()),
                                              median=float(np.ma.median(maskarray)),
                                              majority=float(stats.mode(maskarray, axis=None)[0][0]),count=maskarray.count(),
                                              names=names))
                # print zstats
            else:
                zstats.append(setFeatureStats(id, min=snodata, mean=snodata, max=snodata, sum=snodata, sd=snodata,
                                              median=snodata, majority=snodata, count=snodata, names=names))
        tmpras = None
        tmpds = None
        iter += 1
        if (iter % 100000 == 0):
            print("iter", iter, "of", lyr.GetFeatureCount())
        feat = lyr.GetNextFeature()
    return zstats

def zonalStatisticsDelta(vectorpath, rasterpath, deltapath, deltavalue, deltatype='percent', minvalue=0.0):
    """
    Zonal statistics using a second raster layer to exclude values from statistic calculations. Currently,
    This is implemented as follows. For each zone represented in vectorpath, the median of deltapath is identified.
    Cells from deltapath where the absolute value of ((median - deltapath)/median)*100 is greater than delta value are
    excluded from statistical calculations.

    Args:
        vectorpath: Vector zones
        rasterpath: Raster to calculate statistics from
        deltapath: Raster to exclude cells for statistic calculation
        deltavalue: Threshold for exclusion from statistic calculations
        deltatype: Type of theshold to use. Currently, only percent is available.

    Returns:

    """
    rasterds = raster.openGDALRaster(rasterpath)
    deltads = raster.openGDALRaster(deltapath)
    vectords = vector.openOGRDataSource(vectorpath)
    lyr = vectords.GetLayer()
    geot = rasterds.GetGeoTransform()
    nodata = rasterds.GetRasterBand(1).GetNoDataValue()
    zstats = []
    outofbounds = []
    feat = lyr.GetNextFeature()
    iter = 0
    while feat:
        tmpds = vector.createOGRDataSource('temp', 'Memory')
        tmplyr = tmpds.CreateLayer('polygons', None, ogr.wkbPolygon)
        tmplyr.CreateFeature(feat.Clone())
        offsets = bboxToOffsets(feat.GetGeometryRef().GetEnvelope(), geot)
        for i in range(0, len(offsets)):
            if offsets[i] < 0:
                offsets[i] = 0
        if not any(x < 0 for x in offsets):
            array = rasterds.GetRasterBand(1).ReadAsArray(offsets[2], offsets[0], (offsets[3]-offsets[2]),
                                                          (offsets[1]-offsets[0]))
            deltaarray = deltads.GetRasterBand(1).ReadAsArray(offsets[2], offsets[0], (offsets[3]-offsets[2]),
                                                          (offsets[1]-offsets[0]))
            newgeot = raster.getOffsetGeot(offsets[0], offsets[2], geot)
            tmpras = raster.createGDALRaster('', offsets[1] - offsets[0], offsets[3] - offsets[2], datatype=gdal.GDT_Byte,
                                             drivername='MEM', geot=newgeot)
            if tmpras is not None and array is not None:
                gdal.RasterizeLayer(tmpras, [1], tmplyr, burn_values=[1])
                tmparray = tmpras.ReadAsArray()
                if array.size == np.logical_or(array == nodata, np.logical_not(tmparray)).size:
                    testmaskarray = np.ma.MaskedArray(array,
                                                      mask=np.logical_or(array == nodata, np.logical_not(tmparray)))
                    testmean = np.ma.mean(testmaskarray)

                    if testmean != nodata and deltaarray is not None:
                        deltamaskarray = np.ma.MaskedArray(deltaarray,
                                                           mask=np.logical_or(array == nodata, np.logical_not(tmparray)))
                        median = np.ma.median(deltamaskarray)
                        diff = (abs(deltamaskarray - median) / median) * 100.0
                        maskarray = np.ma.MaskedArray(array, mask=np.logical_or(np.ma.getmask(deltamaskarray),
                                                                                np.logical_or(diff > deltavalue,
                                                                                              deltaarray < minvalue)))
                        #print feat.GetFID(), maskarray.count(), maskarray.min(), maskarray.mean(), maskarray.max()

                        zstats.append(setFeatureStats(feat.GetFID(), min=float(maskarray.min()), mean=float(maskarray.mean()),
                                                      max=float(maskarray.max()), sum=float(maskarray.sum()), sd=float(maskarray.std()),
                                                      median=float(np.ma.median(maskarray)), majority=float(stats.mode(maskarray, axis=None)[0][0]),
                                                      deltamed=float((median*30*30)/1000000), count=maskarray.count()))
                    else:
                        #print "mean = nodata", nodata, testmaskarray
                        zstats.append(setFeatureStats(feat.GetFID()))
                else:
                    #print "array sizes not equal"
                    zstats.append(setFeatureStats(feat.GetFID()))
            else:
                #print "NoneType arrays"
                zstats.append(setFeatureStats(feat.GetFID()))

        else:
            print("out of bounds", feat.GetFID())
            zstats.append(setFeatureStats(feat.GetFID()))
            outofbounds.append(feat.GetFID())
        tmpras = None
        tmpds = None
        iter+=1
        if (iter % 100000 == 0):
            print("iter", iter, "of", lyr.GetFeatureCount())
        feat = lyr.GetNextFeature()
    return zstats

def zonalStatistics_rasterZones(rasterzones, rasterpath, indentifier="fid"):
    print("loading zones")
    zones = raster.getRasterBandAsArray(rasterzones, 1)
    print("loading values")
    rasterds = raster.openGDALRaster(rasterpath)
    rastervals = rasterds.GetRasterBand(1).ReadAsArray()
    nodata = rasterds.GetRasterBand(1).GetNoDataValue()
    print("finding unique values, no data", nodata)
    uvals = np.unique(zones)
    zstats = []
    iter = 0
    print("starting loop", len(uvals))
    for uval in uvals:
        vals = rastervals[np.where(zones == uval)]
        vals = vals[vals != nodata]
        if vals.size > 0:
            zstats.append(setFeatureStats(uval, max=float(vals.max()), min=float(vals.min()),
                                          mean=float(vals.max()), sd=float(vals.std()),
                                          median=float(np.median(vals)),
                                          majority=float(stats.mode(vals, axis=None)[0][0]),
                                          count=vals.size, idname=indentifier))
        else:
            # print "no values for zone", uval, iter
            zstats.append(setFeatureStats(uval, idname=indentifier))
        iter += 1
        if (iter % 100 == 0):
            print("iter", iter, "of", len(uvals))
    return zstats

def zonalStatisticsDelta_methodtest(vectorpath, rasterpath, deltapath, idfield="fid", deltamax=None, deltamin=None, minvalue=0.0):
    """
    Zonal statistics using a second raster layer to exclude values from statistic calculations. Currently,
    This is implemented as follows. For each zone represented in vectorpath, the median of deltapath is identified.
    Cells from deltapath where the absolute value of ((median - deltapath)/median)*100 is greater than delta value are
    excluded from statistical calculations.

    Args:
        vectorpath: Vector zones
        rasterpath: Raster to calculate statistics from
        deltapath: Raster to exclude cells for statistic calculation
        deltavalue: Threshold for exclusion from statistic calculations
        deltatype: Type of theshold to use. Currently, only percent is available.

    Returns:

    """
    np.set_printoptions(suppress=True)
    rasterds = raster.openGDALRaster(rasterpath)
    deltads = raster.openGDALRaster(deltapath)
    vectords = vector.openOGRDataSource(vectorpath)
    lyr = vectords.GetLayer()
    geot = rasterds.GetGeoTransform()
    nodata = rasterds.GetRasterBand(1).GetNoDataValue()
    zstats = []
    outofbounds = []
    feat = lyr.GetNextFeature()
    iter = 0
    while feat:
        if idfield is "fid" or idfield is "FID":
            id = feat.GetFID()
        else:
            id = feat.GetField(idfield)
        tmpds = vector.createOGRDataSource('temp', 'Memory')
        tmplyr = tmpds.CreateLayer('polygons', None, ogr.wkbPolygon)
        tmplyr.CreateFeature(feat.Clone())
        offsets = bboxToOffsets(feat.GetGeometryRef().GetEnvelope(), geot)
        for i in range(0, len(offsets)):
            if offsets[i] < 0:
                offsets[i] = 0
        if not any(x < 0 for x in offsets):
            array = rasterds.GetRasterBand(1).ReadAsArray(offsets[2], offsets[0], (offsets[3]-offsets[2]),
                                                          (offsets[1]-offsets[0]))
            deltaarray = deltads.GetRasterBand(1).ReadAsArray(offsets[2], offsets[0], (offsets[3]-offsets[2]),
                                                              (offsets[1]-offsets[0]))
            newgeot = raster.getOffsetGeot(offsets[0], offsets[2], geot)
            tmpras = raster.createGDALRaster('', offsets[1] - offsets[0], offsets[3] - offsets[2], datatype=gdal.GDT_Byte,
                                             drivername='MEM', geot=newgeot)
            if tmpras is not None and array is not None:
                gdal.RasterizeLayer(tmpras, [1], tmplyr, burn_values=[1])
                tmparray = tmpras.ReadAsArray()
                if array.size == np.logical_or(array == nodata, np.logical_not(tmparray)).size:
                    testmaskarray = np.ma.MaskedArray(array,
                                                      mask=np.logical_or(array == nodata, np.logical_not(tmparray)))
                    testmean = np.ma.mean(testmaskarray)

                    if testmean != nodata and deltaarray is not None:
                        deltamaskarray = np.ma.MaskedArray(deltaarray,
                                                           mask=np.logical_or(array == nodata, np.logical_not(tmparray)))
                        median = np.ma.median(deltamaskarray)

                        ########################################
                        #Difference calculation here
                        ########################################
                        diff = ((deltamaskarray - median) / (deltamaskarray + median))
                        maskarray = np.ma.MaskedArray(array, mask=np.logical_or(np.ma.getmask(deltamaskarray),
                                                                                np.logical_or(np.logical_or(
                                                                                    diff > deltamax,
                                                                                    diff < deltamin),
                                                                                              deltaarray < minvalue)))
                        #print feat.GetFID(), maskarray.count(), maskarray.min(), maskarray.mean(), maskarray.max()

                        zstats.append(setFeatureStats(id, min=float(maskarray.min()), mean=float(maskarray.mean()),
                                                      max=float(maskarray.max()), sum=float(maskarray.sum()), sd=float(maskarray.std()),
                                                      median=float(np.ma.median(maskarray)), majority=float(stats.mode(maskarray, axis=None)[0][0]),
                                                      deltamed=float((median*30*30)/1000000), count=maskarray.count(), idname=idfield))
                        # if id==22877591:
                        #     print "subtract"
                        #     print deltamaskarray-median
                        #     print "add"
                        #     print deltamaskarray + median
                        #     print "buffer"
                        #     print tmparray
                        #     print "fac"
                        #     print deltaarray
                        #     print "diff"
                        #     print diff
                        #     print "values", deltamin, deltamax
                        #     #print np.where((deltaarray > minvalue) & (tmparray == 1) & (diff > deltamin), deltaarray, 0)
                        #     print np.where((deltaarray > minvalue) & (diff < deltamax) & (diff > deltamin), deltaarray, 0)
                        #     print np.where((deltaarray > minvalue) & (diff < deltamax) & (diff > deltamin), diff, 0)
                        #     print id, "count", maskarray.count(), offsets
                    else:
                        #print "mean = nodata", nodata, testmaskarray
                        zstats.append(setFeatureStats(id, idname=idfield))
                else:
                    #print "array sizes not equal"
                    zstats.append(setFeatureStats(id, idname=idfield))
            else:
                #print "NoneType arrays"
                zstats.append(setFeatureStats(id, idname=idfield))

        else:
            print("out of bounds", feat.GetFID())
            zstats.append(setFeatureStats(id))
            outofbounds.append(feat.GetFID())
        tmpras = None
        tmpds = None
        iter+=1
        if (iter % 10000 == 0):
            print("iter", iter, "of", lyr.GetFeatureCount())
        feat = lyr.GetNextFeature()
    return zstats
